import timeit
from typing import Callable
import numpy as np

from diffeq_solver_tk.diffeq import OrdinaryDifferentialEquationMetadata, \
    HeatEquationMetadata, WaveEquationMetadata, BoundaryType


def solve_first_order_ode(metadata: OrdinaryDifferentialEquationMetadata) -> np.ndarray:
    if metadata.samples <= 1 or metadata.time == 0:
        raise Exception("invalid first order ode metadata")

    start_time = timeit.default_timer()
    N = metadata.samples
    dt = metadata.time / (N - 1)
    source: Callable[[float, float], float] = lambda t, x: eval(metadata.source)
    solution = np.zeros(N)
    solution[0] = metadata.initial_derivatives[0]
    for i in range(1, N):
        solution[i] = solution[i - 1] + source(i * dt, solution[i - 1]) * dt
    total_time = timeit.default_timer() - start_time
    print(f"Computing first order ode solution took {total_time} seconds")
    return solution


def solve_second_order_ode(metadata: OrdinaryDifferentialEquationMetadata) -> np.ndarray:
    if metadata.samples <= 1 or metadata.time == 0:
        raise Exception("invalid second order ode metadata")

    start_time = timeit.default_timer()
    N = metadata.samples
    dt = metadata.time / (N - 1)
    initial_value, derivative = metadata.initial_derivatives
    source: Callable[[float, float, float], float] = lambda t, x, y: eval(metadata.source)
    solution = np.zeros(N)
    solution[0] = initial_value
    for i in range(1, N):
        derivative_delta = source((i - 1) * dt, solution[i - 1], derivative) * dt
        derivative = derivative + derivative_delta
        solution[i] = solution[i - 1] + derivative * dt
    total_time = timeit.default_timer() - start_time
    print(f"Computing second order ode solution took {total_time} seconds")
    return solution


def solve_heat_equation(metadata: HeatEquationMetadata) -> np.ndarray:
    if metadata.samples <= 1 or metadata.length == 0 or metadata.time == 0:
        raise Exception("invalid heat equation metadata")

    start_time = timeit.default_timer()
    initial_values: Callable[[float], float] = lambda x: eval(metadata.initial_values)
    source: Callable[[float, float], float] = lambda t, x: eval(metadata.source)
    N = metadata.samples
    dx = metadata.length / (N - 1)
    K = 2 * int(2 * metadata.alpha * metadata.time / dx ** 2) + 1
    dt = metadata.time / (K - 1)
    D = create_heat_equation_time_step_matrix(metadata.alpha * dt / dx ** 2, N)
    left_values: Callable[[float], float] = lambda t: eval(metadata.boundary_conditions.left.values)
    right_values: Callable[[float], float] = lambda t: eval(metadata.boundary_conditions.right.values)
    is_left_dirichlet = metadata.boundary_conditions.left.type == BoundaryType.DIRICHLET
    is_right_dirichlet = metadata.boundary_conditions.right.type == BoundaryType.DIRICHLET

    solution = np.zeros((K, N))
    solution[0] = initial_values(np.arange(N) * dx)
    for k in range(1, K):
        T = k * dt
        solution[k, 1:N - 1] = D @ solution[k - 1]
        solution[k, 1:N - 1] += source(T, np.arange(1, N - 1) * dx) * dt
        solution[k, 0] = left_values(T) if is_left_dirichlet else solution[k, 1] - left_values(T) * dx
        solution[k, -1] = right_values(T) if is_right_dirichlet else solution[k, -2] + right_values(T) * dx

    total_time = timeit.default_timer() - start_time
    print(f"Computing heat equation solution took {total_time} seconds")
    return solution


def solve_wave_equation(metadata: WaveEquationMetadata) -> np.ndarray:
    if metadata.samples <= 1 or metadata.length == 0 or metadata.time == 0:
        raise Exception("invalid wave equation metadata")

    start_time = timeit.default_timer()
    initial_values: Callable[[float], float] = lambda x: eval(metadata.initial_values)
    initial_derivatives: Callable[[float], float] = lambda x: eval(metadata.initial_derivatives)
    source: Callable[[float, float], float] = lambda t, x: eval(metadata.source)
    N = metadata.samples
    dx = metadata.length / (N - 1)
    K = 2 * int(metadata.wave_speed * metadata.time / dx) + 1
    dt = metadata.time / (K - 1)
    D = create_wave_equation_time_step_matrix((metadata.wave_speed * dt / dx) ** 2, N)
    left_values: Callable[[float], float] = lambda t: eval(metadata.boundary_conditions.left.values)
    right_values: Callable[[float], float] = lambda t: eval(metadata.boundary_conditions.right.values)
    is_left_dirichlet = metadata.boundary_conditions.left.type == BoundaryType.DIRICHLET
    is_right_dirichlet = metadata.boundary_conditions.right.type == BoundaryType.DIRICHLET

    # Initialize solution array
    solution = np.zeros((K, N))
    solution[0] = initial_values(np.arange(N) * dx)
    solution[1, 1:N - 1] = (1 / 2) * D @ solution[0]
    solution[1, 1:N - 1] += dt ** 2 * initial_derivatives(np.arange(1, N - 1) * dx)
    solution[1, 1:N - 1] += (dt ** 2 / 2) * source(dt, np.arange(1, N - 1) * dx)
    solution[1, 0] = left_values(dt) if is_left_dirichlet else solution[1, 1] - left_values(dt) * dx
    solution[1, -1] = right_values(dt) if is_right_dirichlet else solution[1, -2] + right_values(dt) * dx
    for k in range(2, K):
        T = k * dt
        solution[k, 1:N - 1] = D @ solution[k - 1] - solution[k - 2, 1:N - 1]
        solution[k, 1:N - 1] += dt ** 2 * source(T, np.arange(1, N - 1) * dx)
        solution[k, 0] = left_values(T) if is_left_dirichlet else solution[k, 1] - left_values(T) * dx
        solution[k, -1] = right_values(T) if is_right_dirichlet else solution[k, -2] + right_values(T) * dx

    total_time = timeit.default_timer() - start_time
    print(f"Computing wave equation solution took {total_time} seconds")
    return solution


def create_heat_equation_time_step_matrix(r: float, N: int) -> np.ndarray:
    D = np.zeros((N - 2, N))
    for i in range(N - 2):
        D[i, i] = r
        D[i, i + 1] = 1 - 2 * r
        D[i, i + 2] = r
    return D


def create_wave_equation_time_step_matrix(r: float, N: int) -> np.ndarray:
    D = np.zeros((N - 2, N))
    for i in range(N - 2):
        D[i, i] = r
        D[i, i + 1] = 2 * (1 - r)
        D[i, i + 2] = r
    return D



from ._differential_equation_form import DifferentialEquationForm
from ._first_order_ode_form import FirstOrderODEForm
from ._second_order_ode_form import SecondOrderODEForm
from ._heat_equation_form import HeatEquationForm
from ._wave_equation_form import WaveEquationForm



app_name = "Differential Equation Solver"
solution_recorded_message = "Your solution has been recorded in {0}"

# Boundary types
dirichlet = "Dirichlet"
neumann = "Neumann"
left_boundary_type = "Left Boundary Type"
right_boundary_type = "Right Boundary Type"

# Button text
solve = "Solve"
export = "Export"
show_plot = "Show Plot"
play = "Play"
pause = "Pause"



import numpy as np
from openpyxl import Workbook


def export_ode_solution(solution, metadata, table_path: str):
    domain = np.linspace(0, metadata.time, metadata.samples)
    workbook = Workbook()
    worksheet = workbook.active
    worksheet.cell(1, 1, 't')
    worksheet.cell(1, 2, 'x')
    row = 2
    for t, x in zip(domain, solution):
        worksheet.cell(row, 1, t)
        worksheet.cell(row, 2, x)
        row += 1
    workbook.save(table_path)


def export_bounded_equation_solution(solution, metadata, table_path: str):
    K, N = solution.shape
    time_domain = np.linspace(0, metadata.time, K)
    space_domain = np.linspace(0, metadata.length, N)
    workbook = Workbook()
    worksheet = workbook.active
    worksheet.cell(1, 2, "x →")
    for i in range(N):
        worksheet.cell(1, 3 + i, space_domain[i])
    worksheet.cell(2, 1, "t ↓")
    for k in range(K):
        worksheet.cell(3 + k, 1, time_domain[k])
    for k in range(K):
        for i in range(N):
            worksheet.cell(k + 3, i + 3, solution[k, i])
    workbook.save(table_path)



import argparse
import json

from diffeq_solver_tk import __version__
from diffeq_solver_tk.diffeq import to_ode_metadata, to_heat_equation_metadata, to_wave_equation_metadata
from diffeq_solver_tk.diffeq import solve_first_order_ode, solve_second_order_ode, solve_heat_equation, \
    solve_wave_equation
from diffeq_solver_tk.excel_export import export_ode_solution, export_bounded_equation_solution


def write_solution(infile: str, outfile: str, equation_type: str):
    with open(infile) as file:
        query = json.load(file)
        match equation_type:
            case "ode1" | "ode2":
                metadata = to_ode_metadata(query)
                solution = solve_first_order_ode(metadata) \
                    if equation_type == "ode1" \
                    else solve_second_order_ode(metadata)
                export_ode_solution(solution, metadata, outfile)
            case "heat":
                metadata = to_heat_equation_metadata(query)
                solution = solve_heat_equation(metadata)
                export_bounded_equation_solution(solution, metadata, outfile)
            case "wave":
                metadata = to_wave_equation_metadata(query)
                solution = solve_wave_equation(metadata)
                export_bounded_equation_solution(solution, metadata, outfile)
            case _:
                raise Exception(f"Invalid equation type {equation_type}")


def main():
    parser = argparse.ArgumentParser(description="Numerically solves differential equations.")
    parser.add_argument("-V", "--version", action="store_true", help="Show version and exit")
    parser.add_argument("type", type=str, choices=["ode1", "ode2", "heat", "wave"], help="DE type to solve")
    parser.add_argument("infile", type=str, help="DE metadata source file path")
    parser.add_argument("outfile", type=str, help="Solution output file path", default="solution.xlsx")
    args = parser.parse_args()

    if args.version:
        print(f"diffeq-solver-tk {__version__}")
        return

    write_solution(args.infile, args.outfile, args.type)


if __name__ == "__main__":
    main()



from typing import TypeVar, Generic, Optional, Callable

import numpy as np
from matplotlib import cm
from matplotlib.animation import FuncAnimation
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from diffeq_solver_tk.diffeq import OrdinaryDifferentialEquationMetadata, BoundedEquationMetadata
from diffeq_solver_tk.excel_export import export_bounded_equation_solution, export_ode_solution

T = TypeVar('T')


class DifferentialEquationService(Generic[T]):
    solve: Callable[[T], np.ndarray]
    main_figure: Optional[Figure]
    metadata: Optional[T] = None
    solution: Optional[np.ndarray] = None

    def __init__(self, solve: Callable[[T], np.ndarray], main_figure: Optional[Figure] = None):
        self.solve = solve
        self.main_figure = main_figure

    def compute_and_update_solution(self, metadata: T):
        self.solution = self.solve(metadata)
        self.metadata = metadata
        self.render_current_solution()

    def clear_solution(self):
        self.metadata = None
        self.solution = None
        self.clear_figure()

    def clear_figure(self):
        self.main_figure.clf()

    def export_solution(self, table_path: str):
        raise NotImplementedError

    def render_current_solution(self):
        raise NotImplementedError


class OrdinaryDifferentialEquationService(DifferentialEquationService[OrdinaryDifferentialEquationMetadata]):
    def render_current_solution(self):
        if self.main_figure is None:
            return
        domain = np.linspace(0, self.metadata.time, self.metadata.samples)
        ax = self.main_figure.add_subplot()
        ax.plot(domain, self.solution)
        ax.set_xlabel("t [time]")
        ax.set_ylabel("x(t)")
        ax.plot()
        ax.set_title("Solution: x(t)")

    def export_solution(self, table_path: str):
        export_ode_solution(self.solution, self.metadata, table_path)


class BoundedEquationService(DifferentialEquationService[BoundedEquationMetadata]):
    __animation: Optional[FuncAnimation] = None
    __is_animation_playing: bool = False

    def render_current_solution(self):
        if self.main_figure is None:
            return
        self.clear_figure()
        self.clear_animation()
        K, N = self.solution.shape
        time_domain = np.linspace(0, self.metadata.time, K)
        space_domain = np.linspace(0, self.metadata.length, N)
        x, t = np.meshgrid(space_domain, time_domain)
        ax = self.main_figure.add_subplot(projection='3d')
        ax.set_xlabel('x [length]')
        ax.set_ylabel('t [time]')
        ax.set_zlabel('u(x, t)')
        surf = ax.plot_surface(x, t, self.solution, cmap=cm.coolwarm, linewidth=0, antialiased=False)
        self.main_figure.colorbar(surf, shrink=0.5, aspect=5)

    def generate_animation(self):
        if self.main_figure is None:
            return
        self.clear_figure()
        K, N = self.solution.shape
        ax = self.main_figure.add_subplot()
        Axes.set_xlim(ax, left=0, right=self.metadata.length)
        Axes.set_ylim(ax, bottom=np.min(self.solution), top=np.max(self.solution))
        ax.set_xlabel("x [length]")
        ax.set_ylabel("u(x, t)")
        ax.set_title("Solution: u(x, t)")
        space_domain = np.linspace(0, self.metadata.length, self.metadata.samples)
        line, = ax.plot(space_domain, self.solution[0], "-o", markersize=4)
        time_text = ax.text(0.82, 0.92, '', transform=ax.transAxes)
        dt = self.metadata.time / K

        def update_plot(k: int):
            line.set_ydata(self.solution[k])
            time_text.set_text("t = %.3f" % (k * dt))
            return line, time_text

        self.__animation = FuncAnimation(self.main_figure, update_plot, frames=K, blit=True, interval=20)

    def toggle_animation(self):
        if not self.__animation:
            self.generate_animation()
            self.__is_animation_playing = True
        elif self.is_animation_playing():
            self.__animation.pause()
            self.__is_animation_playing = False
        else:
            self.__animation.resume()
            self.__is_animation_playing = True

    def is_animation_playing(self):
        return self.__is_animation_playing

    def export_solution(self, table_path: str):
        export_bounded_equation_solution(self.solution, self.metadata, table_path)

    def clear_animation(self):
        if self.main_figure is None or self.__animation is None:
            return
        self.__animation.pause()
        self.__animation = None
        self.__is_animation_playing = False
        self.main_figure.clf()



from dataclasses import dataclass
from enum import Enum
from typing import Union


class BoundaryType(Enum):
    DIRICHLET = "DIRICHLET"
    NEUMANN = "NEUMANN"


@dataclass
class BoundaryCondition:
    type: BoundaryType
    values: str


@dataclass
class BoundaryConditions:
    left: BoundaryCondition
    right: BoundaryCondition


@dataclass
class OrdinaryDifferentialEquationMetadata:
    source: str
    initial_derivatives: list[float]
    samples: int
    time: float


@dataclass
class HeatEquationMetadata:
    boundary_conditions: BoundaryConditions
    length: float
    samples: int
    time: float
    alpha: float
    initial_values: str
    source: str = "0"


@dataclass
class WaveEquationMetadata:
    boundary_conditions: BoundaryConditions
    length: float
    samples: int
    time: float
    wave_speed: float
    initial_values: str
    initial_derivatives: str
    source: str = "0"


BoundedEquationMetadata = Union[HeatEquationMetadata, WaveEquationMetadata]



import tkinter as tk

from diffeq_solver_tk.diffeq import BoundaryConditions, BoundaryCondition, WaveEquationMetadata, BoundaryType, BoundedEquationService
from diffeq_solver_tk.ui.messages import common_messages, wave_equation_messages as messages
from diffeq_solver_tk.ui.forms import DifferentialEquationForm


class WaveEquationForm(DifferentialEquationForm):
    def __init__(self, master: tk.Frame, canvas, equation_service: BoundedEquationService):
        DifferentialEquationForm.__init__(self, master, canvas, equation_service)
        tk.Label(master=self, text=common_messages.left_boundary_type).grid(row=0, column=0, padx=12, pady=6, sticky="w")
        tk.Label(master=self, text=common_messages.right_boundary_type).grid(row=1, column=0, padx=12, pady=6, sticky="w")

        self.left_boundary_type_variable = tk.StringVar(value=BoundaryType.DIRICHLET.value)
        left_dirichlet_button = tk.Radiobutton(master=self, text=common_messages.dirichlet, variable=self.left_boundary_type_variable, value=BoundaryType.DIRICHLET.value)
        left_neumann_button = tk.Radiobutton(master=self, text=common_messages.neumann, variable=self.left_boundary_type_variable, value=BoundaryType.NEUMANN.value)
        left_dirichlet_button.grid(row=0, column=1, sticky="w")
        left_neumann_button.grid(row=0, column=2, sticky="w")

        self.right_boundary_type_variable = tk.StringVar(value=BoundaryType.DIRICHLET.value)
        right_dirichlet_button = tk.Radiobutton(master=self, text=common_messages.dirichlet, variable=self.right_boundary_type_variable, value=BoundaryType.DIRICHLET.value)
        right_neumann_button = tk.Radiobutton(master=self, text=common_messages.neumann, variable=self.right_boundary_type_variable, value=BoundaryType.NEUMANN.value)
        right_dirichlet_button.grid(row=1, column=1, sticky="w")
        right_neumann_button.grid(row=1, column=2, sticky="w")

        self.left_boundary_values_entry = tk.Entry(master=self, width=24)
        self.right_boundary_values_entry = tk.Entry(master=self, width=24)
        self.initial_values_entry = tk.Entry(master=self, width=24)
        self.initial_derivatives_entry = tk.Entry(master=self, width=24)
        self.source_entry = tk.Entry(master=self, width=24)
        self.wave_speed_entry = tk.Entry(master=self, width=24)
        self.length_entry = tk.Entry(master=self, width=24)
        self.time_interval_entry = tk.Entry(master=self, width=24)
        self.samples_entry = tk.Entry(master=self, width=24)
        fields: list[tuple[str, str, tk.Entry]] = [
            (messages.left_boundary_values, messages.left_boundary_values_symbol, self.left_boundary_values_entry),
            (messages.right_boundary_values, messages.right_boundary_values_symbol, self.right_boundary_values_entry),
            (messages.initial_values, messages.initial_values_symbol, self.initial_values_entry),
            (messages.initial_derivatives, messages.initial_derivatives_symbol, self.initial_derivatives_entry),
            (messages.source_term, messages.source_term_symbol, self.source_entry),
            (messages.wave_speed, messages.wave_speed_symbol, self.wave_speed_entry),
            (messages.length, messages.length_symbol, self.length_entry),
            (messages.time_interval, messages.time_interval_symbol, self.time_interval_entry),
            (messages.samples, messages.samples_symbol, self.samples_entry)
        ]

        for i, (label_text, symbol, input_entry) in enumerate(fields):
            row = i + 2
            tk.Label(master=self, text=label_text + ":").grid(row=row, column=0, padx=12, pady=6, sticky="w")
            tk.Label(master=self, text=symbol + " = ").grid(row=row, column=1, padx=0, pady=0, sticky="e")
            input_entry.grid(row=row, column=2, columnspan=2)

        self.solve_button = tk.Button(master=self, text=common_messages.solve, width=10, command=self.solve_equation)
        self.export_button = tk.Button(master=self, text=common_messages.export, width=10, command=self.export_solution)
        self.render_plot_button = tk.Button(master=self, text=common_messages.show_plot, width=10, command=self.handle_render_plot)
        self.toggle_animation_button = tk.Button(master=self, text=common_messages.play, width=10, command=self.handle_toggle_animation)
        self.solve_button.grid(row=11, column=2, pady=6, sticky="w")

    def get_equation_metadata(self):
        left_boundary_type = BoundaryType(self.left_boundary_type_variable.get())
        right_boundary_type = BoundaryType(self.right_boundary_type_variable.get())
        boundary_conditions = BoundaryConditions(
            BoundaryCondition(left_boundary_type,
                              self.left_boundary_values_entry.get()),
            BoundaryCondition(right_boundary_type,
                              self.right_boundary_values_entry.get()))
        source = self.source_entry.get()
        return WaveEquationMetadata(
            boundary_conditions,
            float(self.length_entry.get()),
            int(self.samples_entry.get()),
            float(self.time_interval_entry.get()),
            float(self.wave_speed_entry.get()),
            self.initial_values_entry.get(),
            self.initial_derivatives_entry.get(),
            source if source else "0"
        )

    def on_solve(self):
        self.render_plot_button.grid(row=11, column=3, pady=6, sticky="e")
        self.toggle_animation_button.grid(row=12, column=3, pady=6, sticky="e")
        self.toggle_animation_button.configure(text=common_messages.play)
        self.export_button.grid(row=13, column=3, pady=6, sticky="e")

    def handle_render_plot(self):
        self.toggle_animation_button.configure(text=common_messages.play)
        self.equation_service.render_current_solution()
        self.canvas.draw()

    def handle_toggle_animation(self):
        self.equation_service.toggle_animation()
        self.canvas.draw()
        if self.equation_service.is_animation_playing():
            self.toggle_animation_button.configure(text=common_messages.pause)
        else:
            self.toggle_animation_button.configure(text=common_messages.play)

    def reset(self):
        self.equation_service.clear_solution()
        self.equation_service.clear_animation()
        self.canvas.draw()
        self.render_plot_button.grid_forget()
        self.toggle_animation_button.grid_forget()
        self.export_button.grid_forget()



import numpy as np
from matplotlib.figure import Figure

from diffeq_solver_tk.diffeq import OrdinaryDifferentialEquationMetadata, HeatEquationMetadata, \
    BoundaryConditions, BoundaryCondition, BoundaryType, WaveEquationMetadata, \
    OrdinaryDifferentialEquationService, BoundedEquationService, \
    solve_first_order_ode, solve_second_order_ode, solve_heat_equation, solve_wave_equation


def test_first_order_ode_service():
    # TODO: add more assertions #8
    service = OrdinaryDifferentialEquationService(solve_first_order_ode, Figure())
    metadata = OrdinaryDifferentialEquationMetadata("-4 * x", [1], 50, 4)
    service.compute_and_update_solution(metadata)
    print(service.solution)
    assert service.solution[0] == 1


def test_second_order_ode_service():
    # TODO: add more assertions #8
    service = OrdinaryDifferentialEquationService(solve_second_order_ode, Figure())
    metadata = OrdinaryDifferentialEquationMetadata("-4 * x", [0, 1], 40, 5)
    service.compute_and_update_solution(metadata)
    assert service.solution[0] == 0


def test_heat_equation_service():
    # TODO: add more assertions #8
    service = BoundedEquationService(solve_heat_equation, Figure())
    metadata = HeatEquationMetadata(
        boundary_conditions=BoundaryConditions(
            BoundaryCondition(BoundaryType.NEUMANN, "t"),
            BoundaryCondition(BoundaryType.DIRICHLET, "1")
        ),
        length=1,
        samples=20,
        time=1,
        alpha=1,
        initial_values="0"
    )
    service.compute_and_update_solution(metadata)
    for i in range(20):
        assert service.solution[0, i] == 0


def test_wave_equation_service():
    # TODO: add more assertions #8
    service = BoundedEquationService(solve_wave_equation, Figure())
    metadata = WaveEquationMetadata(
        boundary_conditions=BoundaryConditions(
            BoundaryCondition(BoundaryType.NEUMANN, "t/2"),
            BoundaryCondition(BoundaryType.DIRICHLET, "0")
        ),
        length=1,
        samples=20,
        time=1,
        wave_speed=1,
        initial_values="np.sin(3*x)",
        initial_derivatives="x",
        source="np.exp(-t)*np.sin(x)+1"
    )
    service.compute_and_update_solution(metadata)
    expected = np.sin(3 * np.linspace(0, 1, 20))
    for i in range(20):
        assert service.solution[0, i] == expected[i]



import tkinter as tk

from diffeq_solver_tk.ui.form_container import FormContainer
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure


class App(tk.Frame):
    def __init__(self, master: tk.Tk):
        tk.Frame.__init__(self, master=master)
        figure = Figure(figsize=(6, 1), dpi=91.4)
        canvas = FigureCanvasTkAgg(figure, self)
        canvas.get_tk_widget().pack(side="right", fill="both", expand=True)
        FormContainer(self, canvas).pack(side="left", fill="both")




import tkinter as tk
from diffeq_solver_tk.ui.messages import common_messages
from diffeq_solver_tk.ui.app import App


def main():
    root = tk.Tk(className=common_messages.app_name)
    root.geometry("1100x600")
    App(root).pack(side="top", fill="both", expand=True)
    root.mainloop()


if __name__ == "__main__":
    main()



import tkinter as tk
from tkinter.messagebox import showinfo
from tkinter.filedialog import asksaveasfilename

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

from diffeq_solver_tk.ui.messages import common_messages
from diffeq_solver_tk.diffeq import DifferentialEquationService


class DifferentialEquationForm(tk.Frame):
    canvas: FigureCanvasTkAgg
    equation_service: DifferentialEquationService

    def __init__(self, master: tk.Frame, canvas: FigureCanvasTkAgg, equation_service: DifferentialEquationService):
        tk.Frame.__init__(self, master=master)
        self.canvas = canvas
        self.equation_service = equation_service

    def solve_equation(self):
        try:
            metadata = self.get_equation_metadata()
            self.equation_service.compute_and_update_solution(metadata)
            self.on_solve()
            self.canvas.draw()
        except Exception as err:
            showinfo(common_messages.app_name, str(err))

    def export_solution(self):
        table_path = asksaveasfilename(filetypes=(("Excel files", "*.xlsx"), ("All files", "*.*")))
        if table_path:
            self.equation_service.export_solution(table_path)

    def get_equation_metadata(self):
        raise NotImplementedError

    def on_solve(self):
        raise NotImplementedError

    def reset(self):
        raise NotImplementedError



import tkinter as tk
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

from diffeq_solver_tk.ui.forms import DifferentialEquationForm, FirstOrderODEForm, SecondOrderODEForm, HeatEquationForm, WaveEquationForm
from diffeq_solver_tk.diffeq import OrdinaryDifferentialEquationService, BoundedEquationService, \
    solve_first_order_ode, solve_second_order_ode, solve_heat_equation, solve_wave_equation

class FormContainer(tk.Frame):
    def __init__(self, master: tk.Frame, canvas: FigureCanvasTkAgg):
        tk.Frame.__init__(self, master=master)

        first_order_ode_form = FirstOrderODEForm(self, canvas,
            OrdinaryDifferentialEquationService(solve_first_order_ode, canvas.figure))
        second_order_ode_form = SecondOrderODEForm(self, canvas,
            OrdinaryDifferentialEquationService(solve_second_order_ode, canvas.figure))
        heat_equation_form = HeatEquationForm(self, canvas,
            BoundedEquationService(solve_heat_equation, canvas.figure))
        wave_equation_form = WaveEquationForm(self, canvas,
            BoundedEquationService(solve_wave_equation, canvas.figure))
        self.selected_form = first_order_ode_form

        navbar_frame = tk.Frame(self)
        navbar_frame.pack(side="top", fill="x")
        tk.Button(navbar_frame, text="First Order ODE", command=lambda: self.select_form(first_order_ode_form)).grid(row=0, column=0)
        tk.Button(navbar_frame, text="Second Order ODE", command=lambda: self.select_form(second_order_ode_form)).grid(row=0, column=1)
        tk.Button(navbar_frame, text="Heat Equation", command=lambda: self.select_form(heat_equation_form)).grid(row=0, column=2)
        tk.Button(navbar_frame, text="Wave Equation", command=lambda: self.select_form(wave_equation_form)).grid(row=0, column=3)

        form_frame = tk.Frame(self)
        form_frame.pack(side="bottom", fill="both", expand=True)
        first_order_ode_form.place(in_=form_frame, x=0, relwidth=1, relheight=1)
        second_order_ode_form.place(in_=form_frame, x=0, relwidth=1, relheight=1)
        heat_equation_form.place(in_=form_frame, x=0, relwidth=1, relheight=1)
        wave_equation_form.place(in_=form_frame, x=0, relwidth=1, relheight=1)
        self.selected_form.lift()


    def select_form(self, form: DifferentialEquationForm):
        if self.selected_form != form:
            self.selected_form.reset()
            self.selected_form = form
            self.selected_form.lift()



from . import _common_messages as common_messages
from . import _first_order_ode_messages as first_order_ode_messages
from . import _second_order_ode_messages as second_order_ode_messages
from . import _heat_equation_messages as heat_equation_messages
from . import _wave_equation_messages as wave_equation_messages



import tkinter as tk

from diffeq_solver_tk.diffeq import OrdinaryDifferentialEquationMetadata, OrdinaryDifferentialEquationService
from diffeq_solver_tk.ui.messages import common_messages, first_order_ode_messages as messages
from diffeq_solver_tk.ui.forms import DifferentialEquationForm


class FirstOrderODEForm(DifferentialEquationForm):
    def __init__(self, master: tk.Frame, canvas, equation_service: OrdinaryDifferentialEquationService):
        DifferentialEquationForm.__init__(self, master, canvas, equation_service)
        self.source_entry = tk.Entry(master=self, width=24)
        self.initial_value_entry = tk.Entry(master=self, width=24)
        self.time_entry = tk.Entry(master=self, width=24)
        self.samples_entry = tk.Entry(master=self, width=24)

        fields: list[tuple[str, str, tk.Entry]] = [
            (messages.source_term, messages.source_term_symbol, self.source_entry),
            (messages.initial_value, messages.initial_value_symbol, self.initial_value_entry),
            (messages.time_interval, messages.time_interval_symbol, self.time_entry),
            (messages.samples, messages.samples_symbol, self.samples_entry)
        ]
        for i, (label_text, symbol, input_entry) in enumerate(fields):
            tk.Label(master=self, text=label_text + ":").grid(row=i, column=0, padx=12, pady=6, sticky="w")
            tk.Label(master=self, text=symbol + " = ").grid(row=i, column=1, padx=0, pady=0, sticky="e")
            input_entry.grid(row=i, column=2, columnspan=2)

        tk.Button(master=self, text=common_messages.solve, width=10, command=self.solve_equation).grid(row=4, column=2, pady=10, sticky="w")
        self.export_button = tk.Button(master=self, text=common_messages.export, command=self.export_solution)

    def get_equation_metadata(self):
        return OrdinaryDifferentialEquationMetadata(
            self.source_entry.get(),
            [float(self.initial_value_entry.get())],
            int(self.samples_entry.get()),
            float(self.time_entry.get())
        )

    def on_solve(self):
        self.export_button.grid(row=5, column=2, sticky="w")

    def reset(self):
        self.equation_service.clear_solution()
        self.export_button.grid_forget()
        self.canvas.draw()
